# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import logging
import pytest

from os import path

from swh.loader.package.debian import (
    DebianLoader, download_package, dsc_information, uid_to_person,
    prepare_person, get_package_metadata, extract_package
)
from swh.loader.package.tests.common import check_snapshot, get_stats


logger = logging.getLogger(__name__)


PACKAGE_FILES = {
    'name': 'cicero',
    'version': '0.7.2-3',
    'files': {
        'cicero_0.7.2-3.diff.gz': {
            'md5sum': 'a93661b6a48db48d59ba7d26796fc9ce',
            'name': 'cicero_0.7.2-3.diff.gz',
            'sha256': 'f039c9642fe15c75bed5254315e2a29f9f2700da0e29d9b0729b3ffc46c8971c',  # noqa
            'size': 3964,
            'uri': 'http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-3.diff.gz'  # noqa
        },
        'cicero_0.7.2-3.dsc': {
            'md5sum': 'd5dac83eb9cfc9bb52a15eb618b4670a',
            'name': 'cicero_0.7.2-3.dsc',
            'sha256': '35b7f1048010c67adfd8d70e4961aefd8800eb9a83a4d1cc68088da0009d9a03',  # noqa
            'size': 1864,
            'uri': 'http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-3.dsc'},  # noqa
        'cicero_0.7.2.orig.tar.gz': {
            'md5sum': '4353dede07c5728319ba7f5595a7230a',
            'name': 'cicero_0.7.2.orig.tar.gz',
            'sha256': '63f40f2436ea9f67b44e2d4bd669dbabe90e2635a204526c20e0b3c8ee957786',  # noqa
            'size': 96527,
            'uri': 'http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2.orig.tar.gz'  # noqa
        }
    },
}

PACKAGE_FILES2 = {
    'name': 'cicero',
    'version': '0.7.2-4',
    'files': {
        'cicero_0.7.2-4.diff.gz': {
            'md5sum': '1e7e6fc4a59d57c98082a3af78145734',
            'name': 'cicero_0.7.2-4.diff.gz',
            'sha256': '2e6fa296ee7005473ff58d0971f4fd325617b445671480e9f2cfb738d5dbcd01',  # noqa
            'size': 4038,
            'uri': 'http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-4.diff.gz'  # noqa
        },
        'cicero_0.7.2-4.dsc': {
            'md5sum': '1a6c8855a73b4282bb31d15518f18cde',
            'name': 'cicero_0.7.2-4.dsc',
            'sha256': '913ee52f7093913420de5cbe95d63cfa817f1a1daf997961149501894e754f8b',  # noqa
            'size': 1881,
            'uri': 'http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2-4.dsc'},  # noqa
        'cicero_0.7.2.orig.tar.gz': {
            'md5sum': '4353dede07c5728319ba7f5595a7230a',
            'name': 'cicero_0.7.2.orig.tar.gz',
            'sha256': '63f40f2436ea9f67b44e2d4bd669dbabe90e2635a204526c20e0b3c8ee957786',  # noqa
            'size': 96527,
            'uri': 'http://deb.debian.org/debian/pool/contrib/c/cicero/cicero_0.7.2.orig.tar.gz'  # noqa
        }
    }
}


PACKAGE_PER_VERSION = {
    'stretch/contrib/0.7.2-3': PACKAGE_FILES,
}


PACKAGES_PER_VERSION = {
    'stretch/contrib/0.7.2-3': PACKAGE_FILES,
    'buster/contrib/0.7.2-4': PACKAGE_FILES2,
}


def test_debian_first_visit(
        swh_config, requests_mock_datadir):
    """With no prior visit, load a gnu project ends up with 1 snapshot

    """
    loader = DebianLoader(
        url='deb://Debian/packages/cicero',
        date='2019-10-12T05:58:09.165557+00:00',
        packages=PACKAGE_PER_VERSION)

    actual_load_status = loader.load()
    expected_snapshot_id = '3b6b66e6ee4e7d903a379a882684a2a50480c0b4'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    stats = get_stats(loader.storage)
    assert {
        'content': 42,
        'directory': 2,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 1,  # all artifacts under 1 revision
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': {
            'releases/stretch/contrib/0.7.2-3': {
                'target_type': 'revision',
                'target': '2807f5b3f84368b4889a9ae827fe85854ffecf07',
            }
        },
    }  # different than the previous loader as no release is done

    check_snapshot(expected_snapshot, loader.storage)


def test_debian_first_visit_then_another_visit(
        swh_config, requests_mock_datadir):
    """With no prior visit, load a debian project ends up with 1 snapshot

    """
    url = 'deb://Debian/packages/cicero'
    loader = DebianLoader(
        url=url,
        date='2019-10-12T05:58:09.165557+00:00',
        packages=PACKAGE_PER_VERSION)

    actual_load_status = loader.load()

    expected_snapshot_id = '3b6b66e6ee4e7d903a379a882684a2a50480c0b4'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'deb'

    stats = get_stats(loader.storage)
    assert {
        'content': 42,
        'directory': 2,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 1,  # all artifacts under 1 revision
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': {
            'releases/stretch/contrib/0.7.2-3': {
                'target_type': 'revision',
                'target': '2807f5b3f84368b4889a9ae827fe85854ffecf07',
            }
        },
    }  # different than the previous loader as no release is done

    check_snapshot(expected_snapshot, loader.storage)

    # No change in between load
    actual_load_status2 = loader.load()
    assert actual_load_status2['status'] == 'uneventful'
    origin_visit2 = list(loader.storage.origin_visit_get(url))
    assert origin_visit2[-1]['status'] == 'full'
    assert origin_visit2[-1]['type'] == 'deb'

    stats2 = get_stats(loader.storage)
    assert {
        'content': 42 + 0,
        'directory': 2 + 0,
        'origin': 1,
        'origin_visit': 1 + 1,  # a new visit occurred
        'person': 1,
        'release': 0,
        'revision': 1,
        'skipped_content': 0,
        'snapshot': 1,  # same snapshot across 2 visits
    } == stats2

    urls = [
        m.url for m in requests_mock_datadir.request_history
        if m.url.startswith('http://deb.debian.org')
    ]
    # visited each package artifact twice across 2 visits
    assert len(urls) == len(set(urls))


def test_uid_to_person():
    uid = 'Someone Name <someone@orga.org>'
    actual_person = uid_to_person(uid)

    assert actual_person == {
        'name': 'Someone Name',
        'email': 'someone@orga.org',
        'fullname': uid,
    }


def test_prepare_person():
    actual_author = prepare_person({
        'name': 'Someone Name',
        'email': 'someone@orga.org',
        'fullname': 'Someone Name <someone@orga.org>',
    })

    assert actual_author == {
        'name': b'Someone Name',
        'email': b'someone@orga.org',
        'fullname': b'Someone Name <someone@orga.org>',
    }


def test_download_package(datadir, tmpdir, requests_mock_datadir):
    tmpdir = str(tmpdir)  # py3.5 work around (LocalPath issue)
    all_hashes = download_package(PACKAGE_FILES, tmpdir)
    assert all_hashes == {
        'cicero_0.7.2-3.diff.gz': {
            'checksums': {
                'blake2s256': '08b1c438e70d2474bab843d826515147fa4a817f8c4baaf3ddfbeb5132183f21',  # noqa
                'sha1': '0815282053f21601b0ec4adf7a8fe47eace3c0bc',
                'sha1_git': '834ac91da3a9da8f23f47004bb456dd5bd16fe49',
                'sha256': 'f039c9642fe15c75bed5254315e2a29f9f2700da0e29d9b0729b3ffc46c8971c'  # noqa
            },
            'filename': 'cicero_0.7.2-3.diff.gz',
            'length': 3964},
        'cicero_0.7.2-3.dsc': {
            'checksums': {
                'blake2s256': '8c002bead3e35818eaa9d00826f3d141345707c58fb073beaa8abecf4bde45d2',  # noqa
                'sha1': 'abbec4e8efbbc80278236e1dd136831eac08accd',
                'sha1_git': '1f94b2086fa1142c2df6b94092f5c5fa11093a8e',
                'sha256': '35b7f1048010c67adfd8d70e4961aefd8800eb9a83a4d1cc68088da0009d9a03'  # noqa
            },
            'filename': 'cicero_0.7.2-3.dsc',
            'length': 1864},
        'cicero_0.7.2.orig.tar.gz': {
            'checksums': {
                'blake2s256': '9809aa8d2e2dad7f34cef72883db42b0456ab7c8f1418a636eebd30ab71a15a6',  # noqa
                'sha1': 'a286efd63fe2c9c9f7bb30255c3d6fcdcf390b43',
                'sha1_git': 'aa0a38978dce86d531b5b0299b4a616b95c64c74',
                'sha256': '63f40f2436ea9f67b44e2d4bd669dbabe90e2635a204526c20e0b3c8ee957786'  # noqa
            },
            'filename': 'cicero_0.7.2.orig.tar.gz',
            'length': 96527
        }
    }


def test_dsc_information_ok():
    fname = 'cicero_0.7.2-3.dsc'
    dsc_url, dsc_name = dsc_information(PACKAGE_FILES)

    assert dsc_url == PACKAGE_FILES['files'][fname]['uri']
    assert dsc_name == PACKAGE_FILES['files'][fname]['name']


def test_dsc_information_not_found():
    fname = 'cicero_0.7.2-3.dsc'
    package_files = copy.deepcopy(PACKAGE_FILES)
    package_files['files'].pop(fname)

    dsc_url, dsc_name = dsc_information(package_files)

    assert dsc_url is None
    assert dsc_name is None


def test_dsc_information_too_many_dsc_entries():
    # craft an extra dsc file
    fname = 'cicero_0.7.2-3.dsc'
    package_files = copy.deepcopy(PACKAGE_FILES)
    data = package_files['files'][fname]
    fname2 = fname.replace('cicero', 'ciceroo')
    package_files['files'][fname2] = data

    with pytest.raises(
            ValueError, match='Package %s_%s references several dsc' % (
                package_files['name'], package_files['version'])):
        dsc_information(package_files)


def test_get_package_metadata(requests_mock_datadir, datadir, tmp_path):
    tmp_path = str(tmp_path)  # py3.5 compat.
    package = PACKAGE_FILES

    logger.debug('package: %s', package)

    # download the packages
    all_hashes = download_package(package, tmp_path)

    # Retrieve information from package
    _, dsc_name = dsc_information(package)

    dl_artifacts = [(tmp_path, hashes) for hashes in all_hashes.values()]

    # Extract information from package
    extracted_path = extract_package(dl_artifacts, tmp_path)

    # Retrieve information on package
    dsc_path = path.join(path.dirname(extracted_path), dsc_name)
    actual_package_info = get_package_metadata(
        package, dsc_path, extracted_path)

    logger.debug('actual_package_info: %s', actual_package_info)

    assert actual_package_info == {
        'changelog': {
            'date': '2014-10-19T16:52:35+02:00',
            'history': [
                ('cicero', '0.7.2-2'),
                ('cicero', '0.7.2-1'),
                ('cicero', '0.7-1')
            ],
            'person': {
                'email': 'sthibault@debian.org',
                'fullname': 'Samuel Thibault <sthibault@debian.org>',
                'name': 'Samuel Thibault'
            }
        },
        'maintainers': [
            {
                'email': 'debian-accessibility@lists.debian.org',
                'fullname': 'Debian Accessibility Team '
                '<debian-accessibility@lists.debian.org>',
                'name': 'Debian Accessibility Team'
            },
            {
                'email': 'sthibault@debian.org',
                'fullname': 'Samuel Thibault <sthibault@debian.org>',
                'name': 'Samuel Thibault'
            }
        ],
        'name': 'cicero',
        'version': '0.7.2-3'
    }


def test_debian_multiple_packages(swh_config, requests_mock_datadir):
    url = 'deb://Debian/packages/cicero'
    loader = DebianLoader(
        url=url,
        date='2019-10-12T05:58:09.165557+00:00',
        packages=PACKAGES_PER_VERSION)

    actual_load_status = loader.load()
    expected_snapshot_id = 'defc19021187f3727293121fcf6c5c82cb923604'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'deb'

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': {
            'releases/stretch/contrib/0.7.2-3': {
                'target_type': 'revision',
                'target': '2807f5b3f84368b4889a9ae827fe85854ffecf07',
            },
            'releases/buster/contrib/0.7.2-4': {
                'target_type': 'revision',
                'target': '8224139c274c984147ef4b09aa0e462c55a10bd3',
            }
        },
    }

    check_snapshot(expected_snapshot, loader.storage)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import hashlib
import logging
import pytest

from swh.model.hashutil import hash_to_bytes

from swh.loader.core.loader import BufferedLoader, UnbufferedLoader

from . import BaseLoaderTest


class DummyLoader:
    def cleanup(self):
        pass

    def prepare(self, *args, **kwargs):
        pass

    def fetch_data(self):
        pass

    def store_data(self):
        pass

    def prepare_origin_visit(self, *args, **kwargs):
        origin = self.storage.origin_get(
            self._test_prepare_origin_visit_data['origin'])
        self.origin = origin
        self.origin_url = origin['url']
        self.visit_date = datetime.datetime.utcnow()
        self.visit_type = 'git'
        self.storage.origin_visit_add(self.origin_url, self.visit_date,
                                      self.visit_type)

    def parse_config_file(self, *args, **kwargs):
        return {
            'storage': {
                'cls': 'memory',
                'args': {
                }
            },

            'send_contents': True,
            'send_directories': True,
            'send_revisions': True,
            'send_releases': True,
            'send_snapshot': True,

            'content_packet_size': 2,
            'content_packet_size_bytes': 8,
            'directory_packet_size': 2,
            'revision_packet_size': 2,
            'release_packet_size': 2,

            'content_size_limit': 10000,
        }


class DummyUnbufferedLoader(DummyLoader, UnbufferedLoader):
    pass


class DummyBufferedLoader(DummyLoader, BufferedLoader):
    pass


class DummyBaseLoaderTest(BaseLoaderTest):
    def setUp(self):
        self.loader = self.loader_class(logging_class='dummyloader')
        self.loader.visit_type = 'git'
        # do not call voluntarily super().setUp()
        self.storage = self.loader.storage
        contents = [
            {
                'sha1': '34973274ccef6ab4dfaaf86599792fa9c3fe4689',
                'sha1_git': b'bar1',
                'sha256': b'baz1',
                'blake2s256': b'qux1',
                'status': 'visible',
                'data': b'data1',
                'length': 5,
            },
            {
                'sha1': '61c2b3a30496d329e21af70dd2d7e097046d07b7',
                'sha1_git': b'bar2',
                'sha256': b'baz2',
                'blake2s256': b'qux2',
                'status': 'visible',
                'data': b'data2',
                'length': 5,
            },
        ]
        self.expected_contents = [content['sha1'] for content in contents]
        self.in_contents = contents.copy()
        for content in self.in_contents:
            content['sha1'] = hash_to_bytes(content['sha1'])
        self.in_directories = [
            {'id': hash_to_bytes(id_), 'entries': []}
            for id_ in [
                '44e45d56f88993aae6a0198013efa80716fd8921',
                '54e45d56f88993aae6a0198013efa80716fd8920',
                '43e45d56f88993aae6a0198013efa80716fd8920',
            ]
        ]
        person = {
            'name': b'John Doe',
            'email': b'john.doe@institute.org',
            'fullname': b'John Doe <john.doe@institute.org>'
        }
        rev1_id = b'\x00'*20
        rev2_id = b'\x01'*20
        self.in_revisions = [
            {
                'id': rev1_id,
                'type': 'git',
                'date': 1567591673,
                'committer_date': 1567591673,
                'author': person,
                'committer': person,
                'message': b'msg1',
                'directory': None,
                'synthetic': False,
                'metadata': None,
                'parents': [],
            },
            {
                'id': rev2_id,
                'type': 'hg',
                'date': 1567591673,
                'committer_date': 1567591673,
                'author': person,
                'committer': person,
                'message': b'msg2',
                'directory': None,
                'synthetic': False,
                'metadata': None,
                'parents': [],
            },
        ]
        self.in_releases = [
            {
                'name': b'rel1',
                'id': b'\x02'*20,
                'date': None,
                'author': person,
                'target_type': 'revision',
                'target': rev1_id,
                'message': None,
                'synthetic': False,
            },
            {
                'name': b'rel2',
                'id': b'\x03'*20,
                'date': None,
                'author': person,
                'target_type': 'revision',
                'target': rev2_id,
                'message': None,
                'synthetic': False,
            },
        ]
        self.in_origin = {
            'type': self.loader.visit_type,
            'url': 'http://example.com/',
        }
        self.in_snapshot = {
            'id': b'snap1',
            'branches': {},
        }
        self.in_provider = {
            'provider_name': 'Test Provider',
            'provider_type': 'test_provider',
            'provider_url': 'http://example.org/metadata_provider',
            'metadata': {'working': True},
        }
        self.in_tool = {
            'name': 'Test Tool',
            'version': 'v1.2.3',
            'configuration': {'in_the_Matrix': 'maybe'},
        }

        self.storage.origin_add([self.in_origin])

        # used by prepare_origin_visit() when it gets called
        self.loader._test_prepare_origin_visit_data = {
            'origin': self.in_origin,
        }

    def tearDown(self):
        # do not call voluntarily super().tearDown()
        pass


class CoreUnbufferedLoaderTest(DummyBaseLoaderTest):
    loader_class = DummyUnbufferedLoader

    def test_unbuffered_loader(self):
        self.loader.load()  # initialize the loader

        self.loader.send_contents(self.in_contents[0:1])
        self.loader.send_directories(self.in_directories[0:1])
        self.loader.send_revisions(self.in_revisions[0:1])
        self.loader.send_releases(self.in_releases[0:1])

        self.assertCountContents(1)
        self.assertCountDirectories(1)
        self.assertCountRevisions(1)
        self.assertCountReleases(1)

        self.loader.send_contents(self.in_contents[1:])
        self.loader.send_directories(self.in_directories[1:])
        self.loader.send_revisions(self.in_revisions[1:])
        self.loader.send_releases(self.in_releases[1:])

        self.assertCountContents(len(self.in_contents))
        self.assertCountDirectories(len(self.in_directories))
        self.assertCountRevisions(len(self.in_revisions))
        self.assertCountReleases(len(self.in_releases))


class CoreBufferedLoaderTest(DummyBaseLoaderTest):
    loader_class = DummyBufferedLoader

    def test_buffered_loader(self):
        self.loader.load()  # initialize the loader

        self.loader.maybe_load_contents(self.in_contents[0:1])
        self.loader.maybe_load_directories(self.in_directories[0:1])
        self.loader.maybe_load_revisions(self.in_revisions[0:1])
        self.loader.maybe_load_releases(self.in_releases[0:1])

        self.assertCountContents(0)
        self.assertCountDirectories(0)
        self.assertCountRevisions(0)
        self.assertCountReleases(0)

        self.loader.maybe_load_contents(self.in_contents[1:])
        self.loader.maybe_load_directories(self.in_directories[1:])
        self.loader.maybe_load_revisions(self.in_revisions[1:])
        self.loader.maybe_load_releases(self.in_releases[1:])

        self.assertCountContents(len(self.in_contents))
        self.assertCountDirectories(len(self.in_directories))
        self.assertCountRevisions(len(self.in_revisions))
        self.assertCountReleases(len(self.in_releases))

    def test_directory_cascade(self):
        """Checks that sending a directory triggers sending contents"""
        self.loader.load()  # initialize the loader

        self.loader.maybe_load_contents(self.in_contents[0:1])
        self.loader.maybe_load_directories(self.in_directories)

        self.assertCountContents(1)
        self.assertCountDirectories(len(self.in_directories))

    def test_revision_cascade(self):
        """Checks that sending a revision triggers sending contents and
        directories."""

        self.loader.load()  # initialize the loader

        self.loader.maybe_load_contents(self.in_contents[0:1])
        self.loader.maybe_load_directories(self.in_directories[0:1])
        self.loader.maybe_load_revisions(self.in_revisions)

        self.assertCountContents(1)
        self.assertCountDirectories(1)
        self.assertCountRevisions(len(self.in_revisions))

    def test_release_cascade(self):
        """Checks that sending a release triggers sending revisions,
        contents, and directories."""
        self.loader.load()  # initialize the loader

        self.loader.maybe_load_contents(self.in_contents[0:1])
        self.loader.maybe_load_directories(self.in_directories[0:1])
        self.loader.maybe_load_revisions(self.in_revisions[0:1])
        self.loader.maybe_load_releases(self.in_releases)

        self.assertCountContents(1)
        self.assertCountDirectories(1)
        self.assertCountRevisions(1)
        self.assertCountReleases(len(self.in_releases))

    def test_snapshot_cascade(self):
        """Checks that sending a snapshot triggers sending releases,
        revisions, contents, and directories."""
        self.loader.load()  # initialize the loader

        self.loader.maybe_load_contents(self.in_contents[0:1])
        self.loader.maybe_load_directories(self.in_directories[0:1])
        self.loader.maybe_load_revisions(self.in_revisions[0:1])
        self.loader.maybe_load_releases(self.in_releases[0:1])
        self.loader.maybe_load_snapshot(self.in_snapshot)

        self.assertCountContents(1)
        self.assertCountDirectories(1)
        self.assertCountRevisions(1)
        self.assertCountReleases(1)
        self.assertCountSnapshots(1)

    def test_origin_metadata(self):
        self.loader.load()

        provider_id = self.loader.send_provider(self.in_provider)
        tool_id = self.loader.send_tool(self.in_tool)

        self.loader.send_origin_metadata(
            self.loader.visit_date, provider_id,
            tool_id, {'test_metadata': 'foobar'})

        self.assertOriginMetadataContains(self.in_origin['url'],
                                          {'test_metadata': 'foobar'})

        with self.assertRaises(AssertionError):
            self.assertOriginMetadataContains(self.in_origin['url'],
                                              {'test_metadata': 'foobarbaz'})

        with self.assertRaises(Exception):
            self.assertOriginMetadataContains(self.in_origin['url'] + 'blah',
                                              {'test_metadata': 'foobar'})


def test_loader_logger_default_name():
    loader = DummyBufferedLoader()
    assert isinstance(loader.log, logging.Logger)
    assert loader.log.name == \
        'swh.loader.core.tests.test_loader.DummyBufferedLoader'

    loader = DummyUnbufferedLoader()
    assert isinstance(loader.log, logging.Logger)
    assert loader.log.name == \
        'swh.loader.core.tests.test_loader.DummyUnbufferedLoader'


def test_loader_logger_with_name():
    loader = DummyBufferedLoader('some.logger.name')
    assert isinstance(loader.log, logging.Logger)
    assert loader.log.name == \
        'some.logger.name'


@pytest.mark.fs
def test_loader_save_data_path(tmp_path):
    loader = DummyBufferedLoader('some.logger.name.1')
    url = 'http://bitbucket.org/something'
    loader.origin = {
        'url': url,
    }
    loader.visit_date = datetime.datetime(year=2019, month=10, day=1)
    loader.config = {
        'save_data_path': tmp_path,
    }

    hash_url = hashlib.sha1(url.encode('utf-8')).hexdigest()
    expected_save_path = '%s/sha1:%s/%s/2019' % (
        str(tmp_path), hash_url[0:2], hash_url
    )

    save_path = loader.get_save_data_path()
    assert save_path == expected_save_path



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Convert objects to dictionaries suitable for swh.storage"""

from swh.model.hashutil import hash_to_hex


def content_for_storage(content, log=None, max_content_size=None,
                        origin_url=None):
    """Prepare content to be ready for storage

    Note:
    - 'data' is returned only if max_content_size is not reached.

    Returns:
        content with added data (or reason for being missing)

    """
    ret = content.copy()

    if max_content_size and ret['length'] > max_content_size:
        if log:
            log.info('Skipping content %s, too large (%s > %s)' %
                     (hash_to_hex(content['sha1_git']),
                      ret['length'],
                      max_content_size))
        ret.pop('data', None)
        ret.update({'status': 'absent',
                    'reason': 'Content too large',
                    'origin': origin_url})
        return ret

    if 'data' not in ret:
        ret['data'] = open(ret['path'], 'rb').read()

    # Extra keys added by swh.model.from_disk, that are not accepted
    # by swh-storage
    ret.pop('perms', None)
    ret.pop('path', None)

    ret['status'] = 'visible'

    return ret



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import logging
import tempfile
import os

from typing import (
    Any, Dict, Generator, List, Mapping, Optional, Sequence, Tuple
)

from swh.core.tarball import uncompress
from swh.core.config import SWHConfig
from swh.model.from_disk import Directory
from swh.model.hashutil import hash_to_hex
from swh.model.identifiers import (
    revision_identifier, snapshot_identifier, identifier_to_bytes
)
from swh.storage import get_storage
from swh.storage.algos.snapshot import snapshot_get_all_branches
from swh.loader.core.converters import content_for_storage
from swh.loader.package.utils import download


logger = logging.getLogger(__name__)


# Not implemented yet:
# - clean up disk routines from previous killed workers (when OOMkilled)
# -> separation of concern would like this to be abstracted from the code
# -> experience tells us it's complicated to do as such (T903, T964, T982,
#    etc...)
#
# - model: swh.model.merkle.from_disk should output swh.model.model.* objects
#          to avoid this layer's conversion routine call
# -> Take this up within swh.model's current implementation


class PackageLoader:
    # Origin visit type (str) set by the loader
    visit_type = ''

    def __init__(self, url):
        """Loader's constructor. This raises exception if the minimal required
           configuration is missing (cf. fn:`check` method).

        Args:
            url (str): Origin url to load data from

        """
        # This expects to use the environment variable SWH_CONFIG_FILENAME
        self.config = SWHConfig.parse_config_file()
        self._check_configuration()
        self.storage = get_storage(**self.config['storage'])
        self.url = url
        self.visit_date = datetime.datetime.now(tz=datetime.timezone.utc)

    def _check_configuration(self):
        """Checks the minimal configuration required is set for the loader.

        If some required configuration is missing, exception detailing the
        issue is raised.

        """
        if 'storage' not in self.config:
            raise ValueError(
                'Misconfiguration, at least the storage key should be set')

    def get_versions(self) -> Sequence[str]:
        """Return the list of all published package versions.

        Returns:
            Sequence of published versions

        """
        return []

    def get_package_info(self, version: str) -> Generator[
            Tuple[str, Mapping[str, Any]], None, None]:
        """Given a release version of a package, retrieve the associated
           package information for such version.

        Args:
            version: Package version

        Returns:
            (branch name, package metadata)

        """
        yield from {}

    def build_revision(
            self, a_metadata: Dict, uncompressed_path: str) -> Dict:
        """Build the revision dict from the archive metadata (extrinsic
        artifact metadata) and the intrinsic metadata.

        Args:
            a_metadata: Artifact metadata
            uncompressed_path: Artifact uncompressed path on disk

        Returns:
            SWH data dict

        """
        return {}

    def get_default_version(self) -> str:
        """Retrieve the latest release version if any.

        Returns:
            Latest version

        """
        return ''

    def last_snapshot(self) -> Optional[Dict]:
        """Retrieve the last snapshot

        """
        snapshot = None
        visit = self.storage.origin_visit_get_latest(
            self.url, require_snapshot=True)
        if visit:
            snapshot = snapshot_get_all_branches(
                self.storage, visit['snapshot'])
        return snapshot

    def known_artifacts(self, snapshot: Optional[Dict]) -> Dict:
        """Retrieve the known releases/artifact for the origin.

        Args
            snapshot: snapshot for the visit

        Returns:
            Dict of keys revision id (bytes), values a metadata Dict.

        """
        if not snapshot or 'branches' not in snapshot:
            return {}

        # retrieve only revisions (e.g the alias we do not want here)
        revs = [rev['target']
                for rev in snapshot['branches'].values()
                if rev and rev['target_type'] == 'revision']
        known_revisions = self.storage.revision_get(revs)

        ret = {}
        for revision in known_revisions:
            if not revision:  # revision_get can return None
                continue
            ret[revision['id']] = revision['metadata']

        return ret

    def resolve_revision_from(
            self, known_artifacts: Dict, artifact_metadata: Dict) \
            -> Optional[bytes]:
        """Resolve the revision from a snapshot and an artifact metadata dict.

        If the artifact has already been downloaded, this will return the
        existing revision targeting that uncompressed artifact directory.
        Otherwise, this returns None.

        Args:
            snapshot: Snapshot
            artifact_metadata: Information dict

        Returns:
            None or revision identifier

        """
        return None

    def download_package(self, p_info: Mapping[str, Any],
                         tmpdir: str) -> List[Tuple[str, Mapping]]:
        """Download artifacts for a specific package. All downloads happen in
        in the tmpdir folder.

        Default implementation expects the artifacts package info to be
        about one artifact per package.

        Note that most implementation have 1 artifact per package. But some
        implementation have multiple artifacts per package (debian), some have
        none, the package is the artifact (gnu).

        Args:
            artifacts_package_info: Information on the package artifacts to
                download (url, filename, etc...)
            tmpdir: Location to retrieve such artifacts

        Returns:
            List of (path, computed hashes)

        """
        a_uri = p_info['url']
        filename = p_info.get('filename')
        return [download(a_uri, dest=tmpdir, filename=filename)]

    def uncompress(self, dl_artifacts: List[Tuple[str, Mapping[str, Any]]],
                   dest: str) -> str:
        """Uncompress the artifact(s) in the destination folder dest.

        Optionally, this could need to use the p_info dict for some more
        information (debian).

        """
        uncompressed_path = os.path.join(dest, 'src')
        for a_path, _ in dl_artifacts:
            uncompress(a_path, dest=uncompressed_path)
        return uncompressed_path

    def load(self) -> Dict:
        """Load for a specific origin the associated contents.

        for each package version of the origin

        1. Fetch the files for one package version By default, this can be
           implemented as a simple HTTP request. Loaders with more specific
           requirements can override this, e.g.: the PyPI loader checks the
           integrity of the downloaded files; the Debian loader has to download
           and check several files for one package version.

        2. Extract the downloaded files By default, this would be a universal
           archive/tarball extraction.

           Loaders for specific formats can override this method (for instance,
           the Debian loader uses dpkg-source -x).

        3. Convert the extracted directory to a set of Software Heritage
           objects Using swh.model.from_disk.

        4. Extract the metadata from the unpacked directories This would only
           be applicable for "smart" loaders like npm (parsing the
           package.json), PyPI (parsing the PKG-INFO file) or Debian (parsing
           debian/changelog and debian/control).

           On "minimal-metadata" sources such as the GNU archive, the lister
           should provide the minimal set of metadata needed to populate the
           revision/release objects (authors, dates) as an argument to the
           task.

        5. Generate the revision/release objects for the given version. From
           the data generated at steps 3 and 4.

        end for each

        6. Generate and load the snapshot for the visit

        Using the revisions/releases collected at step 5., and the branch
        information from step 0., generate a snapshot and load it into the
        Software Heritage archive

        """
        status_load = 'uneventful'  # either: eventful, uneventful, failed
        status_visit = 'full'       # either: partial, full
        tmp_revisions = {}  # type: Dict[str, List]
        snapshot = None

        try:
            # Prepare origin and origin_visit
            origin = {'url': self.url}
            self.storage.origin_add_one(origin)
            visit_id = self.storage.origin_visit_add(
                origin=self.url,
                date=self.visit_date,
                type=self.visit_type)['visit']
            last_snapshot = self.last_snapshot()
            logger.debug('last snapshot: %s', last_snapshot)
            known_artifacts = self.known_artifacts(last_snapshot)
            logger.debug('known artifacts: %s', known_artifacts)

            # Retrieve the default release version (the "latest" one)
            default_version = self.get_default_version()
            logger.debug('default version: %s', default_version)

            for version in self.get_versions():  # for each
                logger.debug('version: %s', version)
                tmp_revisions[version] = []
                # `p_` stands for `package_`
                for branch_name, p_info in self.get_package_info(version):
                    logger.debug('package_info: %s', p_info)
                    revision_id = self.resolve_revision_from(
                        known_artifacts, p_info['raw'])
                    if revision_id is None:
                        with tempfile.TemporaryDirectory() as tmpdir:
                            try:
                                dl_artifacts = self.download_package(
                                    p_info, tmpdir)
                            except Exception:
                                logger.exception('Unable to retrieve %s',
                                                 p_info)
                                status_visit = 'partial'
                                continue

                            uncompressed_path = self.uncompress(
                                dl_artifacts, dest=tmpdir)
                            logger.debug('uncompressed_path: %s',
                                         uncompressed_path)

                            directory = Directory.from_disk(
                                path=uncompressed_path.encode('utf-8'),
                                data=True)  # noqa
                            # FIXME: Try not to load the full raw content in
                            # memory
                            objects = directory.collect()

                            contents = objects['content'].values()
                            logger.debug('Number of contents: %s',
                                         len(contents))

                            self.storage.content_add(
                                [content_for_storage(x) for x in contents])

                            status_load = 'eventful'
                            directories = list(objects['directory'].values())

                            logger.debug('Number of directories: %s',
                                         len(directories))

                            self.storage.directory_add(directories)

                            # FIXME: This should be release. cf. D409
                            revision = self.build_revision(
                                p_info['raw'], uncompressed_path)
                            revision.update({
                                'synthetic': True,
                                'directory': directory.hash,
                            })

                        revision['metadata'].update({
                            'original_artifact': [
                                hashes for _, hashes in dl_artifacts
                            ],
                        })

                        revision['id'] = revision_id = identifier_to_bytes(
                            revision_identifier(revision))

                        logger.debug('Revision: %s', revision)

                        self.storage.revision_add([revision])

                    tmp_revisions[version].append((branch_name, revision_id))

            logger.debug('tmp_revisions: %s', tmp_revisions)
            # Build and load the snapshot
            branches = {}  # type: Dict[bytes, Mapping[str, Any]]
            for version, branch_name_revisions in tmp_revisions.items():
                if version == default_version and \
                   len(branch_name_revisions) == 1:
                    # only 1 branch (no ambiguity), we can create an alias
                    # branch 'HEAD'
                    branch_name, _ = branch_name_revisions[0]
                    # except for some corner case (deposit)
                    if branch_name != 'HEAD':
                        branches[b'HEAD'] = {
                            'target_type': 'alias',
                            'target': branch_name.encode('utf-8'),
                        }

                for branch_name, target in branch_name_revisions:
                    branches[branch_name.encode('utf-8')] = {
                        'target_type': 'revision',
                        'target': target,
                    }

            snapshot = {
                'branches': branches
            }
            logger.debug('snapshot: %s', snapshot)

            snapshot['id'] = identifier_to_bytes(
                snapshot_identifier(snapshot))

            logger.debug('snapshot: %s', snapshot)
            self.storage.snapshot_add([snapshot])
            if hasattr(self.storage, 'flush'):
                self.storage.flush()
        except Exception:
            logger.exception('Fail to load %s' % self.url)
            status_visit = 'partial'
            status_load = 'failed'
        finally:
            self.storage.origin_visit_update(
                origin=self.url, visit_id=visit_id, status=status_visit,
                snapshot=snapshot and snapshot['id'])
        result = {
            'status': status_load,
        }  # type: Dict[str, Any]
        if snapshot:
            result['snapshot_id'] = hash_to_hex(snapshot['id'])
        return result



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from os import path

import pytest

from unittest.mock import patch

from swh.core.tarball import uncompress
from swh.core.pytest_plugin import requests_mock_datadir_factory
from swh.model.hashutil import hash_to_bytes

from swh.loader.package.pypi import (
    PyPILoader, pypi_api_url, author, extract_intrinsic_metadata
)
from swh.loader.package.tests.common import (
    check_snapshot, check_metadata_paths, get_stats
)


def test_author_basic():
    data = {
        'author': "i-am-groot",
        'author_email': 'iam@groot.org',
    }
    actual_author = author(data)

    expected_author = {
        'fullname': b'i-am-groot <iam@groot.org>',
        'name': b'i-am-groot',
        'email': b'iam@groot.org',
    }

    assert actual_author == expected_author


def test_author_empty_email():
    data = {
        'author': 'i-am-groot',
        'author_email': '',
    }
    actual_author = author(data)

    expected_author = {
        'fullname': b'i-am-groot',
        'name': b'i-am-groot',
        'email': b'',
    }

    assert actual_author == expected_author


def test_author_empty_name():
    data = {
        'author': "",
        'author_email': 'iam@groot.org',
    }
    actual_author = author(data)

    expected_author = {
        'fullname': b' <iam@groot.org>',
        'name': b'',
        'email': b'iam@groot.org',
    }

    assert actual_author == expected_author


def test_author_malformed():
    data = {
        'author': "['pierre', 'paul', 'jacques']",
        'author_email': None,
    }

    actual_author = author(data)

    expected_author = {
        'fullname': b"['pierre', 'paul', 'jacques']",
        'name': b"['pierre', 'paul', 'jacques']",
        'email': None,
    }

    assert actual_author == expected_author


def test_author_malformed_2():
    data = {
        'author': '[marie, jeanne]',
        'author_email': '[marie@some, jeanne@thing]',
    }

    actual_author = author(data)

    expected_author = {
        'fullname': b'[marie, jeanne] <[marie@some, jeanne@thing]>',
        'name': b'[marie, jeanne]',
        'email': b'[marie@some, jeanne@thing]',
    }

    assert actual_author == expected_author


def test_author_malformed_3():
    data = {
        'author': '[marie, jeanne, pierre]',
        'author_email': '[marie@somewhere.org, jeanne@somewhere.org]',
    }

    actual_author = author(data)

    expected_author = {
        'fullname': b'[marie, jeanne, pierre] <[marie@somewhere.org, jeanne@somewhere.org]>',  # noqa
        'name': b'[marie, jeanne, pierre]',
        'email': b'[marie@somewhere.org, jeanne@somewhere.org]',
    }

    actual_author == expected_author


# configuration error #

def test_badly_configured_loader_raise(monkeypatch):
    """Badly configured loader should raise"""
    monkeypatch.delenv('SWH_CONFIG_FILENAME', raising=False)
    with pytest.raises(ValueError) as e:
        PyPILoader(url='some-url')

    assert 'Misconfiguration' in e.value.args[0]


def test_pypi_api_url():
    """Compute pypi api url from the pypi project url should be ok"""
    url = pypi_api_url('https://pypi.org/project/requests')
    assert url == 'https://pypi.org/pypi/requests/json'


def test_pypi_api_url_with_slash():
    """Compute pypi api url from the pypi project url should be ok"""
    url = pypi_api_url('https://pypi.org/project/requests/')
    assert url == 'https://pypi.org/pypi/requests/json'


@pytest.mark.fs
def test_extract_intrinsic_metadata(tmp_path, datadir):
    """Parsing existing archive's PKG-INFO should yield results"""
    uncompressed_archive_path = str(tmp_path)
    archive_path = path.join(
        datadir, 'https_files.pythonhosted.org', '0805nexter-1.1.0.zip')
    uncompress(archive_path, dest=uncompressed_archive_path)

    actual_metadata = extract_intrinsic_metadata(uncompressed_archive_path)
    expected_metadata = {
        'metadata_version': '1.0',
        'name': '0805nexter',
        'version': '1.1.0',
        'summary': 'a simple printer of nested lest',
        'home_page': 'http://www.hp.com',
        'author': 'hgtkpython',
        'author_email': '2868989685@qq.com',
        'platforms': ['UNKNOWN'],
    }

    assert actual_metadata == expected_metadata


@pytest.mark.fs
def test_extract_intrinsic_metadata_failures(tmp_path):
    """Parsing inexistent path/archive/PKG-INFO yield None"""
    tmp_path = str(tmp_path)  # py3.5 work around (PosixPath issue)
    # inexistent first level path
    assert extract_intrinsic_metadata('/something-inexistent') == {}
    # inexistent second level path (as expected by pypi archives)
    assert extract_intrinsic_metadata(tmp_path) == {}
    # inexistent PKG-INFO within second level path
    existing_path_no_pkginfo = path.join(tmp_path, 'something')
    os.mkdir(existing_path_no_pkginfo)
    assert extract_intrinsic_metadata(tmp_path) == {}


# LOADER SCENARIO #

# "edge" cases (for the same origin) #


# no release artifact:
# {visit full, status: uneventful, no contents, etc...}
requests_mock_datadir_missing_all = requests_mock_datadir_factory(ignore_urls=[
    'https://files.pythonhosted.org/packages/ec/65/c0116953c9a3f47de89e71964d6c7b0c783b01f29fa3390584dbf3046b4d/0805nexter-1.1.0.zip',  # noqa
    'https://files.pythonhosted.org/packages/c4/a0/4562cda161dc4ecbbe9e2a11eb365400c0461845c5be70d73869786809c4/0805nexter-1.2.0.zip',  # noqa
])


def test_no_release_artifact(swh_config, requests_mock_datadir_missing_all):
    """Load a pypi project with all artifacts missing ends up with no snapshot

    """
    url = 'https://pypi.org/project/0805nexter'
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'uneventful'
    assert actual_load_status['snapshot_id'] is not None

    stats = get_stats(loader.storage)
    assert {
        'content': 0,
        'directory': 0,
        'origin': 1,
        'origin_visit': 1,
        'person': 0,
        'release': 0,
        'revision': 0,
        'skipped_content': 0,
        'snapshot': 1,
    } == stats

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'partial'
    assert origin_visit['type'] == 'pypi'


# problem during loading:
# {visit: partial, status: uneventful, no snapshot}


def test_release_with_traceback(swh_config):
    url = 'https://pypi.org/project/0805nexter'
    with patch('swh.loader.package.pypi.PyPILoader.get_default_version',
               side_effect=ValueError('Problem')):
        loader = PyPILoader(url)

        actual_load_status = loader.load()
        assert actual_load_status == {'status': 'failed'}

        stats = get_stats(loader.storage)

        assert {
            'content': 0,
            'directory': 0,
            'origin': 1,
            'origin_visit': 1,
            'person': 0,
            'release': 0,
            'revision': 0,
            'skipped_content': 0,
            'snapshot': 0,
        } == stats

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'partial'
    assert origin_visit['type'] == 'pypi'


# problem during loading: failure early enough in between swh contents...
# some contents (contents, directories, etc...) have been written in storage
# {visit: partial, status: eventful, no snapshot}

# problem during loading: failure late enough we can have snapshots (some
# revisions are written in storage already)
# {visit: partial, status: eventful, snapshot}

# "normal" cases (for the same origin) #


requests_mock_datadir_missing_one = requests_mock_datadir_factory(ignore_urls=[
    'https://files.pythonhosted.org/packages/ec/65/c0116953c9a3f47de89e71964d6c7b0c783b01f29fa3390584dbf3046b4d/0805nexter-1.1.0.zip',  # noqa
])

# some missing release artifacts:
# {visit partial, status: eventful, 1 snapshot}


def test_revision_metadata_structure(swh_config, requests_mock_datadir):
    url = 'https://pypi.org/project/0805nexter'
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    assert actual_load_status['status'] == 'eventful'
    assert actual_load_status['snapshot_id'] is not None

    expected_revision_id = hash_to_bytes(
        'e445da4da22b31bfebb6ffc4383dbf839a074d21')
    revision = list(loader.storage.revision_get([expected_revision_id]))[0]

    assert revision is not None

    check_metadata_paths(revision['metadata'], paths=[
        ('intrinsic.tool', str),
        ('intrinsic.raw', dict),
        ('extrinsic.provider', str),
        ('extrinsic.when', str),
        ('extrinsic.raw', dict),
        ('original_artifact', list),
    ])

    for original_artifact in revision['metadata']['original_artifact']:
        check_metadata_paths(original_artifact, paths=[
            ('filename', str),
            ('length', int),
            ('checksums', dict),
        ])


def test_visit_with_missing_artifact(
        swh_config, requests_mock_datadir_missing_one):
    """Load a pypi project with some missing artifacts ends up with 1 snapshot

    """
    url = 'https://pypi.org/project/0805nexter'
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = 'dd0e4201a232b1c104433741dbf45895b8ac9355'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    stats = get_stats(loader.storage)

    assert {
        'content': 3,
        'directory': 2,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 1,
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    expected_contents = map(hash_to_bytes, [
        '405859113963cb7a797642b45f171d6360425d16',
        'e5686aa568fdb1d19d7f1329267082fe40482d31',
        '83ecf6ec1114fd260ca7a833a2d165e71258c338',
    ])

    assert list(loader.storage.content_missing_per_sha1(expected_contents))\
        == []

    expected_dirs = map(hash_to_bytes, [
        'b178b66bd22383d5f16f4f5c923d39ca798861b4',
        'c3a58f8b57433a4b56caaa5033ae2e0931405338',
    ])

    assert list(loader.storage.directory_missing(expected_dirs)) == []

    # {revision hash: directory hash}
    expected_revs = {
        hash_to_bytes('e445da4da22b31bfebb6ffc4383dbf839a074d21'): hash_to_bytes('b178b66bd22383d5f16f4f5c923d39ca798861b4'),  # noqa
    }
    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_branches = {
        'releases/1.2.0': {
            'target': 'e445da4da22b31bfebb6ffc4383dbf839a074d21',
            'target_type': 'revision',
        },
        'HEAD': {
            'target': 'releases/1.2.0',
            'target_type': 'alias',
        },
    }

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': expected_branches,
    }
    check_snapshot(expected_snapshot, storage=loader.storage)

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'partial'
    assert origin_visit['type'] == 'pypi'


def test_visit_with_1_release_artifact(swh_config, requests_mock_datadir):
    """With no prior visit, load a pypi project ends up with 1 snapshot

    """
    url = 'https://pypi.org/project/0805nexter'
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = 'ba6e158ada75d0b3cfb209ffdf6daa4ed34a227a'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    stats = get_stats(loader.storage)
    assert {
        'content': 6,
        'directory': 4,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 2,
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    expected_contents = map(hash_to_bytes, [
        'a61e24cdfdab3bb7817f6be85d37a3e666b34566',
        '938c33483285fd8ad57f15497f538320df82aeb8',
        'a27576d60e08c94a05006d2e6d540c0fdb5f38c8',
        '405859113963cb7a797642b45f171d6360425d16',
        'e5686aa568fdb1d19d7f1329267082fe40482d31',
        '83ecf6ec1114fd260ca7a833a2d165e71258c338',
    ])

    assert list(loader.storage.content_missing_per_sha1(expected_contents))\
        == []

    expected_dirs = map(hash_to_bytes, [
        '05219ba38bc542d4345d5638af1ed56c7d43ca7d',
        'cf019eb456cf6f78d8c4674596f1c9a97ece8f44',
        'b178b66bd22383d5f16f4f5c923d39ca798861b4',
        'c3a58f8b57433a4b56caaa5033ae2e0931405338',
    ])

    assert list(loader.storage.directory_missing(expected_dirs)) == []

    # {revision hash: directory hash}
    expected_revs = {
        hash_to_bytes('4c99891f93b81450385777235a37b5e966dd1571'): hash_to_bytes('05219ba38bc542d4345d5638af1ed56c7d43ca7d'),  # noqa
        hash_to_bytes('e445da4da22b31bfebb6ffc4383dbf839a074d21'): hash_to_bytes('b178b66bd22383d5f16f4f5c923d39ca798861b4'),  # noqa
    }
    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_branches = {
        'releases/1.1.0': {
            'target': '4c99891f93b81450385777235a37b5e966dd1571',
            'target_type': 'revision',
        },
        'releases/1.2.0': {
            'target': 'e445da4da22b31bfebb6ffc4383dbf839a074d21',
            'target_type': 'revision',
        },
        'HEAD': {
            'target': 'releases/1.2.0',
            'target_type': 'alias',
        },
    }

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': expected_branches,
    }
    check_snapshot(expected_snapshot, loader.storage)

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'pypi'


def test_multiple_visits_with_no_change(swh_config, requests_mock_datadir):
    """Multiple visits with no changes results in 1 same snapshot

    """
    url = 'https://pypi.org/project/0805nexter'
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    snapshot_id = 'ba6e158ada75d0b3cfb209ffdf6daa4ed34a227a'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': snapshot_id,
    }

    stats = get_stats(loader.storage)

    assert {
        'content': 6,
        'directory': 4,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 2,
        'skipped_content': 0,
        'snapshot': 1
    } == stats

    expected_branches = {
        'releases/1.1.0': {
            'target': '4c99891f93b81450385777235a37b5e966dd1571',
            'target_type': 'revision',
        },
        'releases/1.2.0': {
            'target': 'e445da4da22b31bfebb6ffc4383dbf839a074d21',
            'target_type': 'revision',
        },
        'HEAD': {
            'target': 'releases/1.2.0',
            'target_type': 'alias',
        },
    }

    expected_snapshot = {
        'id': snapshot_id,
        'branches': expected_branches,
    }
    check_snapshot(expected_snapshot, loader.storage)

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'pypi'

    actual_load_status2 = loader.load()
    assert actual_load_status2 == {
        'status': 'uneventful',
        'snapshot_id': actual_load_status2['snapshot_id']
    }

    stats2 = get_stats(loader.storage)
    expected_stats2 = stats.copy()
    expected_stats2['origin_visit'] = 1 + 1
    assert expected_stats2 == stats2

    # same snapshot
    actual_snapshot_id = origin_visit['snapshot']
    assert actual_snapshot_id == hash_to_bytes(snapshot_id)


def test_incremental_visit(swh_config, requests_mock_datadir_visits):
    """With prior visit, 2nd load will result with a different snapshot

    """
    url = 'https://pypi.org/project/0805nexter'
    loader = PyPILoader(url)

    visit1_actual_load_status = loader.load()
    visit1_stats = get_stats(loader.storage)
    expected_snapshot_id = 'ba6e158ada75d0b3cfb209ffdf6daa4ed34a227a'
    assert visit1_actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id
    }

    origin_visit1 = next(loader.storage.origin_visit_get(url))
    assert origin_visit1['status'] == 'full'
    assert origin_visit1['type'] == 'pypi'

    assert {
        'content': 6,
        'directory': 4,
        'origin': 1,
        'origin_visit': 1,
        'person': 1,
        'release': 0,
        'revision': 2,
        'skipped_content': 0,
        'snapshot': 1
    } == visit1_stats

    # Reset internal state
    loader._info = None

    visit2_actual_load_status = loader.load()
    visit2_stats = get_stats(loader.storage)

    assert visit2_actual_load_status['status'] == 'eventful'
    expected_snapshot_id2 = '2e5149a7b0725d18231a37b342e9b7c4e121f283'
    assert visit2_actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id2
    }

    visits = list(loader.storage.origin_visit_get(url))
    assert len(visits) == 2
    assert visits[1]['status'] == 'full'
    assert visits[1]['type'] == 'pypi'

    assert {
        'content': 6 + 1,     # 1 more content
        'directory': 4 + 2,   # 2 more directories
        'origin': 1,
        'origin_visit': 1 + 1,
        'person': 1,
        'release': 0,
        'revision': 2 + 1,    # 1 more revision
        'skipped_content': 0,
        'snapshot': 1 + 1,    # 1 more snapshot
    } == visit2_stats

    expected_contents = map(hash_to_bytes, [
        'a61e24cdfdab3bb7817f6be85d37a3e666b34566',
        '938c33483285fd8ad57f15497f538320df82aeb8',
        'a27576d60e08c94a05006d2e6d540c0fdb5f38c8',
        '405859113963cb7a797642b45f171d6360425d16',
        'e5686aa568fdb1d19d7f1329267082fe40482d31',
        '83ecf6ec1114fd260ca7a833a2d165e71258c338',
        '92689fa2b7fb4d4fc6fb195bf73a50c87c030639'
    ])

    assert list(loader.storage.content_missing_per_sha1(expected_contents))\
        == []

    expected_dirs = map(hash_to_bytes, [
        '05219ba38bc542d4345d5638af1ed56c7d43ca7d',
        'cf019eb456cf6f78d8c4674596f1c9a97ece8f44',
        'b178b66bd22383d5f16f4f5c923d39ca798861b4',
        'c3a58f8b57433a4b56caaa5033ae2e0931405338',
        'e226e7e4ad03b4fc1403d69a18ebdd6f2edd2b3a',
        '52604d46843b898f5a43208045d09fcf8731631b',

    ])

    assert list(loader.storage.directory_missing(expected_dirs)) == []

    # {revision hash: directory hash}
    expected_revs = {
        hash_to_bytes('4c99891f93b81450385777235a37b5e966dd1571'): hash_to_bytes('05219ba38bc542d4345d5638af1ed56c7d43ca7d'),  # noqa
        hash_to_bytes('e445da4da22b31bfebb6ffc4383dbf839a074d21'): hash_to_bytes('b178b66bd22383d5f16f4f5c923d39ca798861b4'),  # noqa
        hash_to_bytes('51247143b01445c9348afa9edfae31bf7c5d86b1'): hash_to_bytes('e226e7e4ad03b4fc1403d69a18ebdd6f2edd2b3a'),  # noqa
    }

    assert list(loader.storage.revision_missing(expected_revs)) == []

    expected_branches = {
        'releases/1.1.0': {
            'target': '4c99891f93b81450385777235a37b5e966dd1571',
            'target_type': 'revision',
        },
        'releases/1.2.0': {
            'target': 'e445da4da22b31bfebb6ffc4383dbf839a074d21',
            'target_type': 'revision',
        },
        'releases/1.3.0': {
            'target': '51247143b01445c9348afa9edfae31bf7c5d86b1',
            'target_type': 'revision',
        },
        'HEAD': {
            'target': 'releases/1.3.0',
            'target_type': 'alias',
        },
    }
    expected_snapshot = {
        'id': expected_snapshot_id2,
        'branches': expected_branches,
    }

    check_snapshot(expected_snapshot, loader.storage)

    origin_visit = list(loader.storage.origin_visit_get(url))[-1]
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'pypi'

    urls = [
        m.url for m in requests_mock_datadir_visits.request_history
        if m.url.startswith('https://files.pythonhosted.org')
    ]
    # visited each artifact once across 2 visits
    assert len(urls) == len(set(urls))


# release artifact, no new artifact
# {visit full, status uneventful, same snapshot as before}

# release artifact, old artifact with different checksums
# {visit full, status full, new snapshot with shared history and some new
# different history}

# release with multiple sdist artifacts per pypi "version"
# snapshot branch output is different

def test_visit_1_release_with_2_artifacts(swh_config, requests_mock_datadir):
    """With no prior visit, load a pypi project ends up with 1 snapshot

    """
    url = 'https://pypi.org/project/nexter'
    loader = PyPILoader(url)

    actual_load_status = loader.load()
    expected_snapshot_id = 'a27e638a4dad6fbfa273c6ebec1c4bf320fb84c6'
    assert actual_load_status == {
        'status': 'eventful',
        'snapshot_id': expected_snapshot_id,
    }

    expected_branches = {
        'releases/1.1.0/nexter-1.1.0.zip': {
            'target': '4c99891f93b81450385777235a37b5e966dd1571',
            'target_type': 'revision',
        },
        'releases/1.1.0/nexter-1.1.0.tar.gz': {
            'target': '0bf88f5760cca7665d0af4d6575d9301134fe11a',
            'target_type': 'revision',
        },
    }

    expected_snapshot = {
        'id': expected_snapshot_id,
        'branches': expected_branches,
    }
    check_snapshot(expected_snapshot, loader.storage)

    origin_visit = next(loader.storage.origin_visit_get(url))
    assert origin_visit['status'] == 'full'
    assert origin_visit['type'] == 'pypi'



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch


@patch('swh.loader.package.archive.ArchiveLoader.load')
def test_gnu_loader(
        mock_loader, swh_app, celery_session_worker, swh_config):
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.tasks.LoadArchive',
        (), dict(url='some-url', artifacts=[]))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}


@patch('swh.loader.package.debian.DebianLoader.load')
def test_debian_loader(
        mock_loader, swh_app, celery_session_worker, swh_config):
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.tasks.LoadDebian',
        (), dict(url='some-url', date='some-date', packages={}))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}


@patch('swh.loader.package.deposit.DepositLoader.load')
def test_deposit_loader(
        mock_loader, swh_app, celery_session_worker, swh_config):
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.tasks.LoadDeposit',
        (), dict(url='some-url', deposit_id='some-d-id'))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}


@patch('swh.loader.package.npm.NpmLoader.load')
def test_npm_loader(
        mock_loader, swh_app, celery_session_worker, swh_config):
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.tasks.LoadNpm',
        (), dict(package_name='some-package',
                 package_url='some',
                 package_metadata_url='something'))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}


@patch('swh.loader.package.pypi.PyPILoader.load')
def test_pypi_loader(
        mock_loader, swh_app, celery_session_worker, swh_config):
    mock_loader.return_value = {'status': 'eventful'}

    res = swh_app.send_task(
        'swh.loader.package.tasks.LoadPyPI',
        (), dict(url='some-url'))
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {'status': 'eventful'}
